import { Logger } from '@nestjs/common';
import {
  WebSocketGateway,
  WebSocketServer,
  OnGatewayConnection,
  OnGatewayDisconnect,
} from '@nestjs/websockets';
import { Server, WebSocket } from 'ws';
import { wsAdapter } from '../main.js';
import { WsError } from './auth.guard.js';
import { AuthService } from './auth.service.js';
import { IncomingMessage } from 'http';
import { parse as parseUrl } from 'url';
import { DBAuthToken } from 'lib';

@WebSocketGateway({
  transports: ['websocket'],
  adapter: wsAdapter,
  path: '/auth/ws',
  cors: {
    origin: ['*'],
    methods: ['*'],
    credentials: true,
  },
})
export class AuthGateway implements OnGatewayConnection, OnGatewayDisconnect {
  constructor(private auth: AuthService) {}

  @WebSocketServer()
  server: Server;

  async handleConnection(socket: WebSocket, request: IncomingMessage) {
    Logger.log('WS connection');

    // Goofy authentication handler, guard hooks don't work on this :(
    function closeSocket(
      ...params: ConstructorParameters<typeof WsError>
    ): void {
      socket.send(new WsError(...params).toString());
      socket.close();
    }

    const token = parseUrl(request.url || '', true).query['auth'];
    if (!token || typeof token != 'string') {
      return closeSocket(
        'UNAUTHORIZED',
        'Pass authentication token in `auth` query param',
      );
    }

    let details: DBAuthToken;

    try {
      details = await this.auth.getToken(token, true);
    } catch (e) {
      return closeSocket('UNAUTHORIZED', 'Invalid token');
    }

    if (!details.pending) {
      return closeSocket(
        'UNAUTHORIZED',
        "Don't connect to this route with a validated session token",
      );
    }

    Logger.log('Client successfully connected to auth WS');
    this.auth.registerAuthRequestListener(socket, details);
  }

  handleDisconnect(socket: WebSocket) {
    Logger.log('Client disconnected from auth WS');
    this.auth.unregisterAuthRequestListener(socket);
  }
}
