import {
  BadRequestException,
  Injectable,
  InternalServerErrorException,
  Logger,
  UnauthorizedException,
} from '@nestjs/common';
import {
  AuthRequestInteractionReq,
  DBAuthToken,
  RE_ULID,
  generateAuthToken,
  generateOTP,
} from 'lib';
import { ulid } from 'ulid';
import { lookup as ipLookup } from 'geoip-lite';
import { DatabaseService } from 'src/database/database.service';
import { LoginResponse } from './auth.controller';
import { UserAuditLogType } from 'lib/dist/cjs/types/database/user_audit_log';
import { WebSocket } from 'ws';
import { OnEvent } from '@nestjs/event-emitter';

@Injectable()
export class AuthService {
  constructor(private db: DatabaseService) {}

  private authRequestListeners: { socket: WebSocket; auth: DBAuthToken }[] = [];

  async getToken(token: string, getInvalid = false): Promise<DBAuthToken> {
    const tokenData = await this.db.getDb().authTokens.findOne({
      token: token,
    });

    if (!tokenData) throw new UnauthorizedException('Invalid token');
    if (tokenData.expires && tokenData.expires.getTime() <= Date.now()) {
      throw new UnauthorizedException('Token expired');
    }
    if (tokenData.pending && !getInvalid) {
      throw new UnauthorizedException('Token not activated yet');
    }
    return tokenData;
  }

  async createLoginRequest(user: string, ip: string): Promise<LoginResponse> {
    if (!new RegExp(RE_ULID).test(user)) {
      throw new BadRequestException(
        "The provided user doesn't appear to be a valid user ID",
      );
    }

    const tokenId = ulid();
    const logId = ulid();
    const token = generateAuthToken();
    let otp: string | undefined = undefined;

    // Eliminate possibility of duplicate OTPs
    for (let i = 0; i < 10; i++) {
      const generated = generateOTP();
      const count = await this.db.getDb().authTokens.countDocuments({
        otp: generated,
        user: user,
        pending: true,
      });

      if (!count) {
        otp = generated;
        break;
      }
    }

    if (!otp) {
      throw new InternalServerErrorException('Unable to find a free OTP');
    }

    await this.db.getDb().authTokens.insertOne({
      _id: tokenId,
      user: user,
      token: token,
      otp: otp,
      audit_log_event: logId,
      pending: true,
      expires: new Date(Date.now() + 1000 * 60 * 5),
    });

    const ipData = ipLookup(ip);

    await this.db.getDb().userAuditLog.insertOne({
      _id: logId,
      user: user,
      type: UserAuditLogType.LoginAttempt,
      location: {
        country: ipData?.country,
        region: ipData?.region,
        city: ipData?.city,
      },
    });

    return {
      token: token,
      otp: otp,
    };
  }

  @OnEvent('auth_request')
  async handleAuthConfirm(data: AuthRequestInteractionReq) {
    const listeners = this.authRequestListeners.filter(
      (l) => l.auth._id == data.tokenId,
    );

    Logger.log(`Messaging and closing ${listeners.length} sockets`);

    for (const { socket } of listeners) {
      if (socket.readyState == socket.OPEN) {
        socket.send(JSON.stringify({ action: data.action }, null, 4));
        socket.close(1000);
      }
    }

    this.authRequestListeners = this.authRequestListeners.filter(
      (l) => l.auth._id != data.tokenId,
    );
  }

  registerAuthRequestListener(socket: WebSocket, auth: DBAuthToken) {
    this.authRequestListeners.push({ socket, auth });
  }

  unregisterAuthRequestListener(socket: WebSocket) {
    this.authRequestListeners = this.authRequestListeners.filter(
      (l) => !Object.is(l.socket, socket),
    );
  }
}
