import { NestFactory } from '@nestjs/core';
import { AppModule } from './app.module';
import {
  DocumentBuilder,
  SwaggerDocumentOptions,
  SwaggerModule,
} from '@nestjs/swagger';
import { META } from './meta';
import { checkEnv } from 'lib';
import { config } from 'dotenv';
import { Logger, ValidationPipe } from '@nestjs/common';
import { maxmindUpdater } from './util';
import { WsAdapter } from '@nestjs/platform-ws';
import { writeFile } from 'fs/promises';
import { join as joinPath } from 'path';

config();
if (process.env.NODE_ENV != 'production') {
  Logger.log('$NODE_ENV is not set; Loading .env.dev');
  config({ path: '../.env.dev' });
}
checkEnv(['MONGODB_URI', 'MONGODB_DATABASE', 'REDIS_URI']);

export let wsAdapter: WsAdapter;

async function bootstrap() {
  const app = await NestFactory.create(AppModule, { cors: true });
  app.getHttpAdapter().getInstance().disable('x-powered-by');

  const swaggerConfig = new DocumentBuilder()
    .setTitle('AutoMod API')
    .setDescription('API to interact with AutoMod')
    .setVersion(META.version)
    .addBearerAuth({ type: 'http', name: 'API Authentication' })
    .addServer('http://localhost:3000', 'Development')
    .build();

  const options: SwaggerDocumentOptions = {
    operationIdFactory: (_controllerKey: string, methodKey: string) =>
      methodKey,
  };

  const spec = SwaggerModule.createDocument(app, swaggerConfig, options);
  SwaggerModule.setup('/api', app, spec, {
    customCssUrl:
      'https://cdn.githubraw.com/Amoenus/SwaggerDark/master/SwaggerDark.css',
  });

  wsAdapter = new WsAdapter(app);
  app.useWebSocketAdapter(wsAdapter);
  app.useGlobalPipes(new ValidationPipe({ skipMissingProperties: false }));

  if (process.env.NODE_ENV == 'development') {
    Logger.log('Updating OpenAPI library');
    await writeFile(
      joinPath(__dirname, '..', '..', 'oapilib', 'OpenAPI.json'),
      JSON.stringify(spec, null, 4),
    );
  }

  await app.listen(3000);

  maxmindUpdater();
}

bootstrap();
