import {
    AUTH_TOKEN_LIFETIME,
    Command,
    embed,
    getDmChannel,
    yesNoMessage,
} from "lib";
import { t, tEmbedMsg } from "../../modules/i18n.js";
import { META } from "../../meta.js";
import { decodeTime } from "ulid";
import { logger } from "../../index.js";
import { Channel } from "revolt.js";
import { authRequestsSender } from "../../modules/messaging.js";

export default {
    name: "login",
    aliases: [],
    description: "Approve login requests",
    async run(message, args, client, context) {
        const otp = args.shift();
        if (!otp) {
            await message.reply(
                await tEmbedMsg(
                    message.author,
                    "commands.login.info",
                    "commands.login.title",
                    "INFO",
                    {
                        prefix: context.serverPrefix,
                        url: META.dashboardLoginUrl,
                    },
                ),
            );
            return;
        }

        const tokenData = await client.db.authTokens.findOne({
            otp: otp,
            pending: true,
            user: message.authorId,
            expires: { $gt: new Date(Date.now()) },
        });

        if (!tokenData) {
            await message.reply(
                await tEmbedMsg(
                    message.author,
                    "commands.login.not_found",
                    "commands.login.title",
                    "ERROR",
                ),
            );
            return;
        }

        const auditLogEvent = tokenData.audit_log_event
            ? await client.db.userAuditLog.findOne({
                  _id: tokenData.audit_log_event,
              })
            : null;

        const country = auditLogEvent?.location?.country;
        const region = auditLogEvent?.location?.region;
        const city = auditLogEvent?.location?.city;
        const time = Math.round(decodeTime(tokenData._id) / 1000);

        let targetChannel = message.channel!;
        if (message.server) {
            try {
                targetChannel = await getDmChannel(message.author!, client);
            } catch (e) {
                logger.warn("Failed to open DM channel: " + e);
            }
        }

        const generateDetailsMessage = async (isDM: boolean) =>
            await t("commands.login.details", message.author, {
                country: isDM ? country || "Unknown" : "***",
                region: isDM ? region || "Unknown" : "***",
                city: isDM ? city || "Unknown" : "***",
                time: `<t:${time}:T>`,
            });

        const sendPrompt = (channel: Channel, isDM: boolean) =>
            // eslint-disable-next-line no-async-promise-executor
            new Promise(async (resolve, reject) => {
                const response = await yesNoMessage(
                    channel,
                    client,
                    message.authorId!,
                    await t("commands.login.prompt", message.author, {
                        url: `[${META.dashboardDomain}](${META.dashboardUrl})`,
                    }),
                    await t("commands.login.title", message.author),
                    await t("commands.login.approved", message.author),
                    await t("commands.login.denied", message.author),
                    120,
                    embed(
                        await generateDetailsMessage(isDM),
                        undefined,
                        "INFO",
                    ),
                    resolve,
                );

                try {
                    if (response) {
                        await client.db.authTokens.updateOne(
                            {
                                otp: otp,
                                pending: true,
                                user: message.authorId,
                            },
                            {
                                $set: {
                                    pending: false,
                                    expires: new Date(
                                        Date.now() + AUTH_TOKEN_LIFETIME,
                                    ),
                                    otp: undefined,
                                },
                            },
                        );
                        await authRequestsSender({
                            action: "approved",
                            tokenId: tokenData._id,
                        });
                    } else {
                        await client.db.authTokens.deleteOne({
                            otp: otp,
                            pending: true,
                            user: message.authorId,
                        });
                        await authRequestsSender({
                            action: "denied",
                            tokenId: tokenData._id,
                        });
                    }
                } catch (e) {
                    logger.error("" + e);
                }
            });

        try {
            await sendPrompt(targetChannel, !targetChannel.serverId);

            if (message.channelId != targetChannel.id) {
                await message.reply(
                    await tEmbedMsg(
                        message.author,
                        "commands.login.check_dm",
                        "commands.login.title",
                        "INFO",
                        {
                            dm_url: `/channel/${targetChannel.id}`,
                        },
                    ),
                );
            }
        } catch (e) {
            await sendPrompt(message.channel!, false);
        }
    },
} as Command;
