"use client";

import {
    apiWithCredentials,
    getPendingAuthData,
    setCredentials,
    setPendingAuthData,
} from "@/lib/auth";
import useAuth from "@/lib/useAuth";
import { useBearStore } from "@/lib/useState";
import { useRouter } from "next/navigation";
import { API } from "oapilib";
import { useEffect, useState } from "react";

type AuthWsCallback = (
    result: "approved" | "denied" | "errored",
    error?: any,
) => any;
class AuthWs {
    private ws: WebSocket | null = null;
    private done = false;
    private token: string;
    private cb: AuthWsCallback;

    constructor(token: string, cb: AuthWsCallback) {
        this.token = token;
        this.cb = cb;
        this.connect();
    }

    private connect() {
        console.debug("Connecting to WS");
        if (this.ws) {
            try {
                this.ws.close();
            } catch (e) {}
        }

        let uri = `${
            apiWithCredentials().config.baseURL
        }/auth/ws?auth=${encodeURIComponent(this.token)}`;

        if (uri.startsWith("http://")) uri = uri.replace("http://", "ws://");
        else if (uri.startsWith("https://"))
            uri = uri.replace("https://", "wss://");
        else throw new Error("Invalid URL protocol");

        this.ws = new WebSocket(uri);
        this.ws.onopen = () => {
            console.debug("WS connected!");
        };

        this.ws.onclose = () => {
            if (this.done) {
                console.debug("WS closed, not reconnecting");
            } else {
                console.debug("WS closed, reconnecting in 5 seconds...");

                setTimeout(() => {
                    this.connect();
                }, 5000);
            }
        };

        this.ws.onmessage = (event) => {
            console.debug("Received WS message");
            const message = JSON.parse(event.data);
            console.debug(message);

            if (message.type == "error") {
                this.done = true;
                this.cb("errored", message);
            } else if (message.action) {
                this.done = true;
                if (message.action == "approved") {
                    this.cb("approved");
                } else {
                    this.cb("denied");
                }
            }
        };
    }
}

export default function Auth() {
    const state = useBearStore();
    const router = useRouter();
    const auth = useAuth({ redirect: false });

    const [pageState, setPageState] = useState<
        "input" | "loading" | "listening" | "denied" | "error"
    >("input");
    const [otp, setOtp] = useState("");
    const [token, setToken] = useState("");
    const [ws, setWs] = useState<AuthWs | null>(null);
    const [error, setError] = useState("");
    const [uid, setUid] = useState(" ");

    const wsCallback =
        (token: string): AuthWsCallback =>
        async (res, error) => {
            if (res == "errored") {
                setPageState("error");
                setError(JSON.stringify(error));
            }
            if (res == "approved") {
                setPageState("loading");
                await setCredentials({
                    token: token,
                });
                state.api = apiWithCredentials(token);

                state.auth = await state.api.get("/auth");

                router.push("/dashboard");
            }
            if (res == "denied") {
                setPageState("denied");
            }
            setOtp("");
            setToken("");
            setWs(null);
            setUid("");

            await setPendingAuthData(null);
        };

    useEffect(() => {
        if (auth?.authenticated) {
            router.push("/dashboard");
        }

        if (pageState != "input") return;

        getPendingAuthData().then(async (data) => {
            if (data) {
                const api = apiWithCredentials(data.token);

                // If this succeeds, we're authenticated. If not, create WS listener instead.
                try {
                    await api.get("/auth");
                    wsCallback(data.token)("approved");
                } catch (_e) {
                    setOtp(data.otp);
                    setToken(data.token);
                    setPageState("listening");
                    setWs(new AuthWs(data.token, wsCallback(data.token)));
                }
            }
        });
    });

    switch (pageState) {
        case "input":
            return (
                <>
                    <h1>Let's get you logged in.</h1>
                    <label htmlFor="uid">Enter your Revolt ID: </label>
                    <input
                        type="text"
                        id="uid"
                        value={uid}
                        onChange={(e) => setUid(e.currentTarget.value)}
                    />{" "}
                    <button
                        onClick={async () => {
                            if (!uid) return;

                            // Let's use a fresh API instance just in case the current one still has authentication set
                            const api = apiWithCredentials();

                            setPageState("loading");

                            try {
                                const result = await api.post("/auth/login", {
                                    user: uid.trim(),
                                });

                                setPageState("listening");
                                setOtp(result.otp);
                                setToken(result.token);
                                setWs(
                                    new AuthWs(
                                        result.token,
                                        wsCallback(result.token),
                                    ),
                                );

                                await setPendingAuthData({
                                    token: result.token,
                                    otp: result.otp,
                                });
                            } catch (e) {
                                console.error(e);
                                try {
                                    setError(JSON.stringify(e, null, 4));
                                } catch (e) {
                                    setError("" + e);
                                }
                                setPageState("error");
                            }
                        }}
                    >
                        Go!
                    </button>
                </>
            );
        case "loading":
            return (
                <>
                    <span>Just one second...</span>
                </>
            );
        case "listening":
            return (
                <>
                    <h1>Just one more step</h1>
                    <p>
                        To finish, run the following in AutoMod's direct
                        messages: <code>/login {otp}</code>
                    </p>
                </>
            );
        case "error":
            return (
                <>
                    <h1>Hm, something went wrong.</h1>
                    <p>{error}</p>
                </>
            );
        case "denied":
            return (
                <>
                    <h1>Authentication request was denied.</h1>
                    <button onClick={() => setPageState("input")}>
                        Try again?
                    </button>
                </>
            );
    }
}
